@@ -26,6 +26,7 @@
 #include <sys/types.h>
 
 #include <pulse/sample.h>
+#include <pulse/format.h>
 #include <pulse/channelmap.h>
 #include <pulse/volume.h>
 #include <pulse/def.h>
@@ -154,7 +155,7 @@
  * index. The application writes to the write index and the sound
  * device reads from the read index. The read index is increased
  * monotonically, while the write index may be freely controlled by
- * the application. Substracting the read index from the write index
+ * the application. Subtracting the read index from the write index
  * will give you the current fill level of the buffer. The read/write
  * indexes are 64bit values and measured in bytes, they will never
  * wrap. The current read/write index may be queried using
@@ -193,7 +194,7 @@
  * The transfer buffers can be controlled through a number of operations:
  *
  * \li pa_stream_cork() - Start or stop the playback or recording.
- * \li pa_stream_trigger() - Start playback immediatly and do not wait for
+ * \li pa_stream_trigger() - Start playback immediately and do not wait for
  *                           the buffer to fill up to the set trigger level.
  * \li pa_stream_prebuf() - Reenable the playback trigger level.
  * \li pa_stream_drain() - Wait for the playback buffer to go empty. Will
@@ -266,7 +267,7 @@
  * pa_stream_get_time() and pa_stream_get_latency() will try to
  * interpolate the current playback time/latency by estimating the
  * number of samples that have been played back by the hardware since
- * the last regular timing update. It is espcially useful to combine
+ * the last regular timing update. It is especially useful to combine
  * this option with PA_STREAM_AUTO_TIMING_UPDATE, which will enable
  * you to monitor the current playback time/latency very precisely and
  * very frequently without requiring a network round trip every time.
@@ -279,7 +280,7 @@
  * pa_stream_set_overflow_callback() and
  * pa_stream_set_underflow_callback().
  *
- * \section sync_streams Sychronizing Multiple Playback Streams
+ * \section sync_streams Synchronizing Multiple Playback Streams
  *
  * PulseAudio allows applications to fully synchronize multiple
  * playback streams that are connected to the same output device. That
@@ -289,7 +290,7 @@
  * issued on the others.
  *
  * To synchronize a stream to another, just pass the "master" stream
- * as last argument to pa_stream_connect_playack(). To make sure that
+ * as last argument to pa_stream_connect_playback(). To make sure that
  * the freshly created stream doesn't start playback right-away, make
  * sure to pass PA_STREAM_START_CORKED and - after all streams have
  * been created - uncork them all with a single call to
@@ -356,6 +357,17 @@ pa_stream* pa_stream_new_with_proplist(
         const pa_channel_map *map         /**< The desired channel map, or NULL for default */,
         pa_proplist *p                    /**< The initial property list */);
 
+/* Create a new, unconnected stream with the specified name, the set of formats
+ * this client can provide, and an initial list of properties. While
+ * connecting, the server will select the most appropriate format which the
+ * client must then provide. \since 1.0 */
+pa_stream *pa_stream_new_extended(
+        pa_context *c                     /**< The context to create this stream in */,
+        const char *name                  /**< A name for this stream */,
+        pa_format_info * const * formats  /**< The list of formats that can be provided */,
+        unsigned int n_formats            /**< The number of formats being passed in */,
+        pa_proplist *p                    /**< The initial property list */);
+
 /** Decrease the reference counter by one */
 void pa_stream_unref(pa_stream *s);
 
@@ -407,7 +419,7 @@ int pa_stream_is_corked(pa_stream *s);
 /** Connect the stream to a sink. It is strongly recommended to pass
  * NULL in both dev and volume and not to set either
  * PA_STREAM_START_MUTED nor PA_STREAM_START_UNMUTED -- unless these
- * options are directly dependant on user input or configuration. If
+ * options are directly dependent on user input or configuration. If
  * you follow this rule then the sound server will have the full
  * flexibility to choose the device, volume and mute status
  * automatically, based on server-side policies, heuristics and stored
@@ -447,7 +459,7 @@ int pa_stream_disconnect(pa_stream *s);
  * can place the data to write and the maximum number of bytes you can
  * write. On return *nbytes can be smaller or have the same value as
  * you passed in. You need to be able to handle both cases. Accessing
- * memory beyond the returned *nbytes value is invalid. Acessing the
+ * memory beyond the returned *nbytes value is invalid. Accessing the
  * memory returned after the following pa_stream_write() or
  * pa_stream_cancel_write() is invalid. On invocation only *nbytes
  * needs to be initialized, on return both *data and *nbytes will be
@@ -558,6 +570,12 @@ void pa_stream_set_read_callback(pa_stream *p, pa_stream_request_cb_t cb, void *
 /** Set the callback function that is called when a buffer overflow happens. (Only for playback streams) */
 void pa_stream_set_overflow_callback(pa_stream *p, pa_stream_notify_cb_t cb, void *userdata);
 
+/** Return at what position the latest underflow occurred, or -1 if this information is not
+ * known (e g if no underflow has occurred, or server is older than 1.0).
+ * Can be used inside the underflow callback to get information about the current underflow.
+ * (Only for playback streams) \since 1.0 */
+int64_t pa_stream_get_underflow_index(pa_stream *p);
+
 /** Set the callback function that is called when a buffer underflow happens. (Only for playback streams) */
 void pa_stream_set_underflow_callback(pa_stream *p, pa_stream_notify_cb_t cb, void *userdata);
 
@@ -590,11 +608,11 @@ void pa_stream_set_moved_callback(pa_stream *p, pa_stream_notify_cb_t cb, void *
  * 0.9.8. \since 0.9.8 */
 void pa_stream_set_suspended_callback(pa_stream *p, pa_stream_notify_cb_t cb, void *userdata);
 
-/** Set the callback function that is called whenver a meta/policy
+/** Set the callback function that is called whenever a meta/policy
  * control event is received.\since 0.9.15 */
 void pa_stream_set_event_callback(pa_stream *p, pa_stream_event_cb_t cb, void *userdata);
 
-/** Set the callback function that is called whenver the buffer
+/** Set the callback function that is called whenever the buffer
  * attributes on the server side change. Please note that the buffer
  * attributes can change when moving a stream to a different
  * sink/source too, hence if you use this callback you should use
@@ -636,7 +654,7 @@ pa_operation* pa_stream_set_name(pa_stream *s, const char *name, pa_stream_succe
  * pa_stream_get_timing_info().
  *
  * This function will usually only return new data if a timing info
- * update has been recieved. Only if timing interpolation has been
+ * update has been received. Only if timing interpolation has been
  * requested (PA_STREAM_INTERPOLATE_TIMING) the data from the last
  * timing update is used for an estimation of the current
  * playback/recording time based on the local time that passed since
@@ -660,7 +678,7 @@ pa_operation* pa_stream_set_name(pa_stream *s, const char *name, pa_stream_succe
  * pa_stream_update_timing_info().
  *
  * If no timing information has been
- * recieved yet this call will return PA_ERR_NODATA. For more details
+ * received yet this call will return PA_ERR_NODATA. For more details
  * see pa_stream_get_timing_info(). */
 int pa_stream_get_time(pa_stream *s, pa_usec_t *r_usec);
 
@@ -671,7 +689,7 @@ int pa_stream_get_time(pa_stream *s, pa_usec_t *r_usec);
  * negative, i.e. the captured samples are not yet played. In this
  * case *negative is set to 1.
  *
- * If no timing information has been recieved yet this call will
+ * If no timing information has been received yet this call will
  * return PA_ERR_NODATA. For more details see
  * pa_stream_get_timing_info() and pa_stream_get_time(). */
 int pa_stream_get_latency(pa_stream *s, pa_usec_t *r_usec, int *negative);
@@ -689,7 +707,7 @@ int pa_stream_get_latency(pa_stream *s, pa_usec_t *r_usec, int *negative);
  *
  * Please note that the write_index member field (and only this field)
  * is updated on each pa_stream_write() call, not just when a timing
- * update has been recieved. */
+ * update has been received. */
 const pa_timing_info* pa_stream_get_timing_info(pa_stream *s);
 
 /** Return a pointer to the stream's sample specification. */
@@ -698,8 +716,11 @@ const pa_sample_spec* pa_stream_get_sample_spec(pa_stream *s);
 /** Return a pointer to the stream's channel map. */
 const pa_channel_map* pa_stream_get_channel_map(pa_stream *s);
 
+/** Return a pointer to the stream's format \since 1.0 */
+const pa_format_info* pa_stream_get_format_info(pa_stream *s);
+
 /** Return the per-stream server-side buffer metrics of the
- * stream. Only valid after the stream has been connected successfuly
+ * stream. Only valid after the stream has been connected successfully
  * and if the server is at least PulseAudio 0.9. This will return the
  * actual configured buffering metrics, which may differ from what was
  * requested during pa_stream_connect_record() or
@@ -720,7 +741,7 @@ pa_operation *pa_stream_set_buffer_attr(pa_stream *s, const pa_buffer_attr *attr
 
 /** Change the stream sampling rate during playback. You need to pass
  * PA_STREAM_VARIABLE_RATE in the flags parameter of
- * pa_stream_connect() if you plan to use this function. Only valid
+ * pa_stream_connect_...() if you plan to use this function. Only valid
  * after the stream has been connected successfully and if the server
  * is at least PulseAudio 0.9.8. \since 0.9.8 */
 pa_operation *pa_stream_update_sample_rate(pa_stream *s, uint32_t rate, pa_stream_success_cb_t cb, void *userdata);
